/* librist. Copyright 2019-2020 SipRadius LLC. All right reserved.
 * Author: Daniele Lacamera <root@danielinux.net>
 * Author: Kuldeep Singh Dhaka <kuldeep@madresistor.com>
 * Author: Sergio Ammirata <sergio@ammirata.net>
 * Author: Gijs Peskens <gijs@in2inip.nl>
 */

#include "rist-private.h"
#include "log-private.h"
#include "udp-private.h"
#include <string.h>
#include "cjson/cJSON.h"

static double round_two_digits(double number)
{
	long new_number = (long)(number * 100);
	return (double)(new_number) / 100;
}

void rist_sender_peer_statistics(struct rist_peer *peer)
{
	// TODO: print warning here?? stale flow?
	if (!peer->authenticated)
	{
		return;
	}

	struct rist_stats *stats_container = malloc(sizeof(struct rist_stats));
	stats_container->stats_type = RIST_STATS_SENDER_PEER;
	stats_container->version = RIST_STATS_VERSION;

	peer->stats_sender_total.received += peer->stats_sender_instant.received;

	size_t retry_buf_size = rist_get_sender_retry_queue_size(peer->sender_ctx);

	struct rist_bandwidth_estimation *cli_bw = &peer->bw;
	struct rist_bandwidth_estimation *retry_bw = &peer->retry_bw;
	// Refresh stats value just in case
	rist_calculate_bitrate_sender(0, cli_bw);
	rist_calculate_bitrate_sender(0, retry_bw);

	double Q = 100;
	if (peer->stats_sender_instant.sent > 0)
	{
		Q = (double)((peer->stats_sender_instant.sent) * 100.0) /
			(double)(peer->stats_sender_instant.sent + peer->stats_sender_instant.bloat_skip + peer->stats_sender_instant.retrans_skip + peer->stats_sender_instant.retrans);
		Q = round_two_digits(Q);
	}

	uint32_t time_left = 0;
	if (peer->sender_ctx->cooldown_time > 0)
	{
		time_left = (uint32_t)(timestampNTP_u64() - peer->sender_ctx->cooldown_time) / 1000;
	}

	uint32_t avg_rtt = (peer->eight_times_rtt / 8);

	struct rist_common_ctx *cctx = get_cctx(peer);

	cJSON *stats = cJSON_CreateObject();
	cJSON *rist_sender_stats = cJSON_AddObjectToObject(stats, "sender-stats");
	cJSON *peer_obj = cJSON_AddObjectToObject(rist_sender_stats, "peer");
	cJSON_AddNumberToObject(peer_obj, "flow_id", peer->adv_flow_id);
	cJSON_AddNumberToObject(peer_obj, "id", peer->adv_peer_id);
	cJSON_AddStringToObject(peer_obj, "cname", peer->receiver_name);
	cJSON_AddStringToObject(peer_obj, "type", peer->is_data ? "data" : "rtcp");
	cJSON *json_stats = cJSON_AddObjectToObject(peer_obj, "stats");
	cJSON_AddNumberToObject(json_stats, "quality", Q);
	cJSON_AddNumberToObject(json_stats, "sent", (double)peer->stats_sender_instant.sent);
	cJSON_AddNumberToObject(json_stats, "received", (double)peer->stats_sender_instant.received);
	cJSON_AddNumberToObject(json_stats, "retransmitted", (double)peer->stats_sender_instant.retrans);
	cJSON_AddNumberToObject(json_stats, "bandwidth", (double)cli_bw->bitrate);
	cJSON_AddNumberToObject(json_stats, "retry_bandwidth", (double)retry_bw->bitrate);
	cJSON_AddNumberToObject(json_stats, "bloat_skipped", (double)peer->stats_sender_instant.bloat_skip);
	cJSON_AddNumberToObject(json_stats, "retransmit_skipped", (double)peer->stats_sender_instant.retrans_skip);
	cJSON_AddNumberToObject(json_stats, "rtt", (double)peer->last_mrtt);
	cJSON_AddNumberToObject(json_stats, "avg_rtt", (double)avg_rtt);
	cJSON_AddNumberToObject(json_stats, "retry_buffer_size", (double)retry_buf_size);
	cJSON_AddNumberToObject(json_stats, "cooldown_time", (double)time_left);
	char *stats_string = cJSON_PrintUnformatted(stats);
	cJSON_Delete(stats);

	stats_container->stats_json = stats_string;
	stats_container->json_size = sizeof(stats_string);

	strncpy(stats_container->stats.sender_peer.cname, peer->receiver_name, RIST_MAX_STRING_SHORT);
	stats_container->stats.sender_peer.peer_id = peer->adv_peer_id;
	stats_container->stats.sender_peer.bandwidth = cli_bw->bitrate;
	// TODO: stats_sender_peer->retry_bandwidth;
	stats_container->stats.sender_peer.sent = peer->stats_sender_instant.sent;
	stats_container->stats.sender_peer.received = peer->stats_sender_instant.received;
	stats_container->stats.sender_peer.retransmitted = peer->stats_sender_instant.retrans;
	stats_container->stats.sender_peer.quality = Q;
	stats_container->stats.sender_peer.rtt = avg_rtt;

	if (cctx->stats_callback != NULL)
		cctx->stats_callback(cctx->stats_callback_argument, stats_container);
	else
		rist_stats_free(stats_container);

	memset(&peer->stats_sender_instant, 0, sizeof(peer->stats_sender_instant));
}

void rist_receiver_flow_statistics(struct rist_receiver *ctx, struct rist_flow *flow)
{
	if (!flow)
		return;

	//Log errors that used to be packet
	if (flow->stats_instant.dropped_full)
		rist_log_priv(&ctx->common, RIST_LOG_ERROR, "Dropped %u packets due to buffers being full\n", flow->stats_instant.dropped_full );
	if (flow->stats_instant.dropped_late)
		rist_log_priv(&ctx->common, RIST_LOG_ERROR, "Dropped %u late packets\n", flow->stats_instant.dropped_late);
        if (flow->stats_instant.lost)
		rist_log_priv(&ctx->common, RIST_LOG_ERROR, "Lost %u packets\n", flow->stats_instant.lost);

	struct rist_stats *stats_container = malloc(sizeof(struct rist_stats));
	stats_container->stats_type = RIST_STATS_RECEIVER_FLOW;
	stats_container->version = RIST_STATS_VERSION;

	if (flow->stats_instant.avg_count)
	{
		flow->stats_instant.cur_ips = (flow->stats_instant.total_ips / flow->stats_instant.avg_count);
	}

	uint64_t flow_recv_instant = 0;
	uint64_t flow_sent_instant = 0;
	uint32_t flow_missing_instant = 0;
	uint32_t flow_recovered_instant = 0;
	uint32_t flow_retries_instant = 0;
	uint32_t flow_dups_instant = 0;
	uint32_t flow_recovered_0nack_instant = 0;
	uint32_t flow_recovered_1nack_instant = 0;
	uint32_t flow_recovered_2nack_instant = 0;
	uint32_t flow_recovered_3nack_instant = 0;
	uint32_t flow_recovered_morenack_instant = 0;
	uint32_t flow_reordered_instant = 0;

	cJSON *stats = cJSON_CreateObject();
	cJSON *stats_obj = cJSON_AddObjectToObject(stats, "receiver-stats");
	cJSON *flow_obj = cJSON_AddObjectToObject(stats_obj, "flowinstant");
	cJSON_AddNumberToObject(flow_obj, "flow_id", flow->flow_id);
	cJSON_AddNumberToObject(flow_obj, "dead",  flow->dead);
	cJSON *json_stats = cJSON_AddObjectToObject(flow_obj, "stats");
	cJSON *peers = cJSON_AddArrayToObject(flow_obj, "peers");
	uint32_t flow_bitrate = 0;
	uint32_t flow_rtt = 0;
	for (size_t i = 0; i < flow->peer_lst_len; i++)
	{
		struct rist_peer *peer = flow->peer_lst[i];
		if (!peer->is_data)
			peer = peer->peer_data;
		uint32_t avg_rtt = (peer->eight_times_rtt / 8);
		uint32_t bitrate;
		size_t eight_times_bitrate;

		bitrate = (uint32_t)peer->bw.bitrate;
		eight_times_bitrate = peer->bw.eight_times_bitrate;

		flow_bitrate += bitrate;
		flow_rtt =+ peer->eight_times_rtt / 8;

		double QpeerInstant = 100;
		if (peer->stats_receiver_instant.recv > 0)
		{
			QpeerInstant = (double)((peer->stats_receiver_instant.recv) * 100.0) /
						   (double)(peer->stats_receiver_instant.recv + peer->stats_receiver_instant.missing);
			QpeerInstant = round_two_digits(QpeerInstant);
		}

		if ((peer->stats_receiver_instant.recovered - peer->stats_receiver_instant.reordered) > 0)
		{
			peer->stats_receiver_instant.recovered_average =
				(peer->stats_receiver_instant.recovered_sum * 100) /
				(peer->stats_receiver_instant.recovered - peer->stats_receiver_instant.reordered);
		}
		else
		{
			peer->stats_receiver_instant.recovered_average = 100;
		}

		peer->stats_receiver_instant.recovered_slope =
			peer->stats_receiver_instant.recovered_3nack -
			peer->stats_receiver_instant.recovered_0nack;

		if ((int32_t)(peer->stats_receiver_instant.recovered_1nack - peer->stats_receiver_instant.recovered_0nack) > 0 &&
			peer->stats_receiver_instant.recovered_1nack != 0 && peer->stats_receiver_instant.recovered_0nack != 0)
		{
			peer->stats_receiver_instant.recovered_slope_inverted++;
		}

		if ((int32_t)(peer->stats_receiver_instant.recovered_2nack - peer->stats_receiver_instant.recovered_1nack) > 0 &&
			peer->stats_receiver_instant.recovered_2nack != 0 && peer->stats_receiver_instant.recovered_1nack != 0)
		{
			peer->stats_receiver_instant.recovered_slope_inverted++;
		}

		if ((int32_t)(peer->stats_receiver_instant.recovered_3nack - peer->stats_receiver_instant.recovered_2nack) > 0 &&
			peer->stats_receiver_instant.recovered_3nack != 0 && peer->stats_receiver_instant.recovered_2nack != 0)
		{
			peer->stats_receiver_instant.recovered_slope_inverted++;
		}

		// Calculate peer totals
		peer->stats_receiver_total.recovered_average = peer->stats_receiver_instant.recovered_average +
													   peer->stats_receiver_total.recovered_average -
													   (peer->stats_receiver_total.recovered_average / 8);

		cJSON *peer_obj = cJSON_CreateObject();
		cJSON_AddNumberToObject(peer_obj, "id", peer->adv_peer_id);
		cJSON_AddNumberToObject(peer_obj, "dead", peer->dead);
		cJSON *peer_stats = cJSON_AddObjectToObject(peer_obj, "stats");
		cJSON_AddNumberToObject(peer_stats, "received", (double)peer->stats_receiver_instant.recv);
		cJSON_AddNumberToObject(peer_stats, "retries", (double)peer->stats_receiver_instant.retries);
		cJSON_AddNumberToObject(peer_stats, "recovery_buffer_length", (double)peer->recovery_buffer_ticks / RIST_CLOCK);
		cJSON_AddNumberToObject(peer_stats, "rtt", (double)peer->last_mrtt);
		cJSON_AddNumberToObject(peer_stats, "avg_rtt", (double)avg_rtt);
		cJSON_AddNumberToObject(peer_stats, "bitrate", (double)bitrate);
		cJSON_AddNumberToObject(peer_stats, "avg_bitrate", (double)eight_times_bitrate / 8);
		cJSON_AddItemToArray(peers, peer_obj);

		// Calculate flow instant stats
		flow_recv_instant += flow->stats_instant.received;
		flow_sent_instant += peer->stats_receiver_instant.sent;
		flow_missing_instant += peer->stats_receiver_instant.missing;
		flow_recovered_instant += peer->stats_receiver_instant.recovered;
		flow_retries_instant += peer->stats_receiver_instant.retries;
		flow_dups_instant += peer->stats_receiver_instant.dups;
		flow_recovered_0nack_instant += peer->stats_receiver_instant.recovered_0nack;
		flow_recovered_1nack_instant += peer->stats_receiver_instant.recovered_1nack;
		flow_recovered_2nack_instant += peer->stats_receiver_instant.recovered_2nack;
		flow_recovered_3nack_instant += peer->stats_receiver_instant.recovered_3nack;
		flow_recovered_morenack_instant += peer->stats_receiver_instant.recovered_morenack;
		flow_reordered_instant += peer->stats_receiver_instant.reordered;

		// buffer_bloat protection flags
		if (peer->config.congestion_control_mode != RIST_CONGESTION_CONTROL_MODE_OFF)
		{
			if (peer->stats_receiver_instant.recovered_slope_inverted >= 3)
			{
				if (!peer->buffer_bloat_active)
				{
					rist_log_priv(&ctx->common, RIST_LOG_INFO,
						"\tActivating buffer protection for peer %d, avg_slope=%d, avg_inverted=%d (%u/%u)\n",
						peer->adv_peer_id,
						peer->stats_receiver_instant.recovered_slope,
						peer->stats_receiver_instant.recovered_slope_inverted,
						peer->stats_receiver_instant.recovered_average,
						peer->stats_receiver_total.recovered_average / 8);
					peer->buffer_bloat_active = true;
				}
			}
			else if (peer->stats_receiver_instant.recovered_slope_inverted == 0)
			{
				if (peer->buffer_bloat_active)
				{
					rist_log_priv(&ctx->common, RIST_LOG_INFO,
						"\tDeactivating buffer protection for peer %d, avg_slope=%d, avg_inverted=%d (%u/%u)\n",
						peer->adv_peer_id,
						peer->stats_receiver_instant.recovered_slope,
						peer->stats_receiver_instant.recovered_slope_inverted,
						peer->stats_receiver_instant.recovered_average,
						peer->stats_receiver_total.recovered_average / 8);
					peer->buffer_bloat_active = false;
				}
			}
		}

		// Clear peer instant stats
		memset(&peer->stats_receiver_instant, 0, sizeof(peer->stats_receiver_instant));
	}

	double Q = 100;
	if (flow_recv_instant > 0)
	{
		Q = (double)((flow_recv_instant)*100.0) /
			(double)(flow_recv_instant + flow_missing_instant);
		Q = round_two_digits(Q);
	}

	// This last one should trigger buffer protection immediately
	if ((flow->missing_counter == 0 || flow_recovered_instant == 0 ||
		 (flow_recovered_instant * 10) < flow_missing_instant) &&
		flow_recv_instant > 10 &&
		flow_recv_instant < flow_missing_instant)
	{
		rist_log_priv(&ctx->common, RIST_LOG_INFO, "\tThe flow link is dead %" PRIu32 " > %" PRIu64 ", deleting all missing queue elements!\n",
			flow_missing_instant, flow_recv_instant);
		/* Delete all missing queue elements (if any) */
		rist_flush_missing_flow_queue(flow);
	}

	uint64_t avg_buffer_duration = 0;
	if (flow->stats_instant.buffer_duration_count > 0)
	{

		for (size_t i = 0; i < flow->stats_instant.buffer_duration_count; i++)
		{
			avg_buffer_duration += flow->stats_instant.buffer_duration[i];
		}
		avg_buffer_duration /= flow->stats_instant.buffer_duration_count;
		flow->stats_instant.buffer_duration_count = 0;
	}
	cJSON_AddNumberToObject(json_stats, "quality", Q);
	cJSON_AddNumberToObject(json_stats, "received", (double)flow_recv_instant);
	cJSON_AddNumberToObject(json_stats, "dropped_late", (double)flow->stats_instant.dropped_late);
	cJSON_AddNumberToObject(json_stats, "dropped_full", (double)flow->stats_instant.dropped_late);
	cJSON_AddNumberToObject(json_stats, "missing", (double)flow_missing_instant);
	cJSON_AddNumberToObject(json_stats, "recovered_total", (double)flow_recovered_instant);
	cJSON_AddNumberToObject(json_stats, "reordered", (double)flow_reordered_instant);
	cJSON_AddNumberToObject(json_stats, "retries", (double)flow_retries_instant);
	cJSON_AddNumberToObject(json_stats, "recovered_one_nack", (double)flow_recovered_0nack_instant);
	cJSON_AddNumberToObject(json_stats, "recovered_two_nacks", (double)flow_recovered_1nack_instant);
	cJSON_AddNumberToObject(json_stats, "recovered_three_nacks", (double)flow_recovered_2nack_instant);
	cJSON_AddNumberToObject(json_stats, "recovered_four_nacks", (double)flow_recovered_3nack_instant);
	cJSON_AddNumberToObject(json_stats, "recovered_more_nacks", (double)flow_recovered_morenack_instant);
	cJSON_AddNumberToObject(json_stats, "lost", (double)flow->stats_instant.lost);
	cJSON_AddNumberToObject(json_stats, "avg_buffer_time", (double)avg_buffer_duration);
	cJSON_AddNumberToObject(json_stats, "duplicates", (double)flow->stats_instant.dupe);
	cJSON_AddNumberToObject(json_stats, "missing_queue", (double)flow->missing_counter);
	cJSON_AddNumberToObject(json_stats, "missing_queue_max", (double)flow->missing_counter_max);
	cJSON_AddNumberToObject(json_stats, "min_inter_packet_spacing", (double)flow->stats_instant.min_ips);
	cJSON_AddNumberToObject(json_stats, "cur_inter_packet_spacing", (double)flow->stats_instant.cur_ips);
	cJSON_AddNumberToObject(json_stats, "max_inter_packet_spacing", (double)flow->stats_instant.max_ips);

	char *stats_string = cJSON_PrintUnformatted(stats);
	cJSON_Delete(stats);

	stats_container->stats_json = stats_string;
	stats_container->json_size = sizeof(stats_string);

	stats_container->stats.receiver_flow.peer_count = (uint32_t)flow->peer_lst_len;
	// TODO: populate stats_receiver_flow->cname
	stats_container->stats.receiver_flow.flow_id = flow->flow_id;
	stats_container->stats.receiver_flow.status = flow->dead;
	stats_container->stats.receiver_flow.bandwidth = flow->peer_lst_len ? flow_bitrate / flow->peer_lst_len : 0;
	//TODO: populate retry_bandwidth;
	stats_container->stats.receiver_flow.sent = flow->peer_lst_len ? flow_sent_instant / flow->peer_lst_len : 0;
	stats_container->stats.receiver_flow.received = flow->peer_lst_len ? flow_recv_instant / flow->peer_lst_len : 0;
	stats_container->stats.receiver_flow.missing = flow_missing_instant;
	stats_container->stats.receiver_flow.reordered = flow_reordered_instant;
	stats_container->stats.receiver_flow.recovered = flow_recovered_instant;
	stats_container->stats.receiver_flow.recovered_one_retry = flow_recovered_0nack_instant;
	stats_container->stats.receiver_flow.lost = flow->stats_instant.lost;
	stats_container->stats.receiver_flow.quality = Q;
	stats_container->stats.receiver_flow.min_inter_packet_spacing = flow->stats_instant.min_ips;
	stats_container->stats.receiver_flow.cur_inter_packet_spacing = flow->stats_instant.cur_ips;
	stats_container->stats.receiver_flow.max_inter_packet_spacing = flow->stats_instant.max_ips;
	stats_container->stats.receiver_flow.rtt = flow->peer_lst_len ? flow_rtt / flow->peer_lst_len : 0;

	/* CALLBACK CALL */
	if (ctx->common.stats_callback != NULL)
		ctx->common.stats_callback(ctx->common.stats_callback_argument, stats_container);
	else
		rist_stats_free(stats_container);

	memset(&flow->stats_instant, 0, sizeof(flow->stats_instant));
	flow->stats_instant.min_ips = 0xFFFFFFFFFFFFFFFFULL;

}
